unit uROR_Errors;

interface

uses
  Classes, SysUtils;

type

/////////////////////////////////// Return Codes \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

  TReturnCode = (
    RC_OK,                      // Successful completion
    RC_DELETE,                  // Delete the current record

    //--- Warnings
    RCW_WARNING,                // Generic Warning

    RCW_CANNOT_LOCK,            // The requested record cannot be locked
    RCW_DELETED,                // Record has been deleted
    RCW_INACTIVATED,            // Record has been inactivated
    RCW_REACTIVATED,            // Record has been reactivated
    RCW_READ_ONLY,              // Read-only data cannot be modified
    RCW_REDUNDANT,              // The action was redunadant

    //--- Errors
    RCE_ERROR,                  // Generic error

    RCE_INVALID_DATA,           // Invalid or missing data
    RCE_INVALID_INPUT,          // Invalid or missing input values
    RCE_INVALID_PTID,           // Invalid or missing patient ID (IEN)
    RCE_INVALID_PARAMS,         // Ivalid value(s) of function parameter(s)
    RCE_LOAD_INACT_REASONS,     // Cannot load inactivation reasons
    RCE_LOAD_PTDATA,            // Cannot load patient data
    RCE_LOAD_PTLABS,            // Cannot load patient's lab results
    RCE_LOAD_SUP_EVIDENCES,     // Cannot load supporting evidences
    RCE_OBJECT_INTEGRITY,       // Object integrity error (missing pointers, etc.)
    RCE_LOAD_AVAIL_REPORTS,     // Cannot load report descriptors
    RCE_RPC_ERROR               // RPC returned an error
  );

////////////////////////////////// TLockError \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

  TLockError = class(TPersistent)
  private
    fDate:     String;
    fJob:      String;
    fPrefix:   String;
    fTask:     String;
    fUserIEN:  String;
    fUserName: String;

  private
    function getMessage: String;

  public
    constructor Create(const RawData: String; const aPrefix: String = '');

    function Show(const AllowReadOnly: Boolean = True): TReturnCode;

    property Date: String                       read fDate;
    property Job: String                        read fJob;
    property Message: String                    read getMessage;
    property Task: String                       read fTask;
    property UserIEN: String                    read fUserIEN;
    property UserName: String                   read fUserName;

  end;

//////////////////////////////////// Exceptions \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

EIntegrityError = class(Exception);

EInvalidPatientIEN = class(EIntegrityError)
public
  constructor Create;
end;

EInvalidRegistryIEN = class(EIntegrityError)
public
  constructor Create;
end;

EMissingPatientDataReference = class(EIntegrityError)
public
  constructor Create;
end;

//////////////////////////// Functions and Procedures \\\\\\\\\\\\\\\\\\\\\\\\\\

  function  IsRCError(rc: TReturnCode): Boolean;
  function  IsRCWarning(rc: TReturnCode): Boolean;

implementation

uses
  uROR_Utilities, Dialogs, uROR_Strings, Controls;

//////////////////////////// Functions and Procedures \\\\\\\\\\\\\\\\\\\\\\\\\\

function IsRCError(rc: TReturnCode): Boolean;
begin
  Result := (rc >= RCE_ERROR);
end;

function IsRCWarning(rc: TReturnCode): Boolean;
begin
  Result := (rc >= RCW_WARNING) and (rc < RCE_ERROR);
end;

//////////////////////////////////// Exceptions \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

constructor EInvalidPatientIEN.Create;
begin
  inherited Create('Inavalid or missing Patient IEN (DFN).');
end;

constructor EInvalidRegistryIEN.Create;
begin
  inherited Create('Inavalid or missing Registry IEN.');
end;

constructor EMissingPatientDataReference.Create;
begin
  inherited Create('Missing reference to patient data.');
end;

////////////////////////////////// TLockError \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

constructor TLockError.Create(const RawData: String; const aPrefix: String);
begin
  inherited Create;
  fDate     := Piece(RawData, '^', 1);
  fUserName := Piece(RawData, '^', 2);
  fUserIEN  := Piece(RawData, '^', 3);
  fJob      := Piece(RawData, '^', 4);
  fTask     := Piece(RawData, '^', 5);

  if aPrefix <> '' then
    fPrefix := aPrefix
  else
    fPrefix := 'Data';
end;

function TLockError.getMessage: String;
begin
  Result := fPrefix + ' is locked' + #13 + 'by ' + UserName;
  if UserIEN <> '' then Result := Result + ' (IEN: ' + UserIEN + ')';
  Result := Result + #13 + 'on ' + FMDateTimeStr(Date) + #13 + 'Job#: ' + Job;
  if Task <> '' then Result := Result + ' Task#: ' + Task;
  Result := Result + '.';
end;

function TLockError.Show(const AllowReadOnly: Boolean): TReturnCode;
var
  mrc: Integer;
begin
  if StrToFloatDef(Date, 0) > 0 then
    begin
      Result := RCW_CANNOT_LOCK;
      if AllowReadOnly then
        begin
          mrc := MessageDlg508('', Message + #13#13 + rscContinueRO,
            mtConfirmation, [mbYes, mbNo], 0);
          if mrc = mrYes then
            Result := RCW_READ_ONLY;
        end
      else
        MessageDlg508('', Message, mtWarning, [mbOK], 0);
    end
  else
    Result := RC_OK;
end;

end.
