unit uROR_HelpMgr;
{
================================================================================
*
*       Package:        ROR - Clinical Case Registries
*       Date Created:   07/27/2014 3:28 pm
*       Site:           Hines OIFO
*       Developers:     Ed Micyus
*
*       Description:    Help File Manager - used to set application help for the
*                       CCR appplication. If CCR is running from a remote drive
*                       the application help files are copied down to the user's
*                       workstation so that context sensitive help works.
*
*       Notes:          Based upon code supplied by Chris Bell (HP)
*
================================================================================
}

interface

uses
  Classes, WinApi.Messages, WinApi.windows, sysutils, WinApi.ActiveX, controls, Forms, StdCtrls,
  Dialogs, WinApi.ShlObj, WinApi.SHFolder, WinApi.ShellAPI, uROR_Strings, uROR_Utilities;

type

  //============= THREADS =============
  tHelpThread = class(TThread)      //Used to copy in the background
  private
    fLocalDir: string;
    fNetDir: string;
    fPatchNum: string;
    fHelpName: string;
  protected
    procedure Execute; override;
  public
    constructor Create(PatchNum, LocalDir, NetDir, HelpName: string);
    destructor Destroy; override;
  published
  end;

//============= PUBLIC functions and procedures =============
procedure LoadHelpFile(AppName, HelpName, FullVersion: string; Count: integer);

//============= PRIVATE functions and procedures =============
function CopyHelpFiles(PatchNum, NetDir, LocalDir: string): Boolean;
function GetLocalAppFolder: string;
function GetOperatingSystem: Integer;
function IsRemoteDrive(FileTocheck: string): Boolean;

const
  fmtCheckName = 'CCR_Patch%s.txt';
  wbemFlagForwardOnly = $00000020;
  wbemFlagReturnImmediately = $00000010;
  cOsUnknown = -1;
  cOsWin95   = 0;
  cOsWin98   = 1;
  cOsWin98SE = 2;
  cOsWinME   = 3;
  cOsWinNT   = 4;
  cOsWin2000 = 5;
  cOsXP      = 6;
  cOSWin7    = 7;

implementation

//============= PUBLIC functions and procedures =============

procedure LoadHelpFile(AppName, HelpName, FullVersion: string; Count: integer);
var
  AppExeName, AppHelpName, AppLocalDir, NetDir, PatchNum, tmpStr: string;
  HelpThread: tHelpThread;
begin
  Screen.Cursor := crHourGlass;
  AppExeName  := Application.ExeName;
  AppHelpName := ExtractFilePath(AppExeName) + HelpName;

  //--- Check if application is running from a server or network install
  if IsRemoteDrive(AppHelpName) then
  begin
    PatchNum := piece(FullVersion, '.', 3);
    AppLocalDir := GetLocalAppFolder + '\' + AppName + '\';
    NetDir := ExtractFilePath(AppExeName);

    //--- Check if the application help has already been copied
    if FileExists(AppLocalDir + Format(fmtCheckName, [PatchNum])) then
      Application.HelpFile := AppLocalDir + HelpName
    else if Count = 0 then
    begin
      tmpStr := StringReplace(fmtCheckName, '%s', '', [rfReplaceAll]);
      if not SysUtils.FileExists(NetDir + tmpStr) then
      begin
        Screen.Cursor := crDefault;
        MessageDlg508('', RSC01002, mtError, [mbOK], 0);
      end
      else begin
        try
          HelpThread := tHelpThread.Create(PatchNum, AppLocalDir, NetDir, HelpName);
          HelpThread.Resume;
        finally
        end;
      end;
     end;
  end
  else
    Application.HelpFile := AppHelpName;

  Screen.Cursor := crDefault;
end;

//============= PRIVATE functions and procedures =============

function GetLocalAppFolder: string;
const
  SHGFP_TYPE_CURRENT = 0;
var
  path: array[0..MaxChar] of char;
begin
  SHGetFolderPath(0, CSIDL_LOCAL_APPDATA, 0, SHGFP_TYPE_CURRENT, @path[0]);
  Result := StrPas(path);
end;

function GetOperatingSystem: Integer;
var
  osVerInfo: TOSVersionInfo;
  majorVer, minorVer: Integer;
begin
  Result := cOsUnknown;
  { set operating system type flag }
  osVerInfo.dwOSVersionInfoSize := SizeOf(TOSVersionInfo);
  if GetVersionEx(osVerInfo) then
  begin
    majorVer := osVerInfo.dwMajorVersion;
    minorVer := osVerInfo.dwMinorVersion;
    case osVerInfo.dwPlatformId of
    VER_PLATFORM_WIN32_NT: { Windows NT/2000 }
      begin
        if majorVer <= 6 then
          Result := cOsWin7;
        if (majorVer = 4) or (majorVer = 5) then
          Result := cOsWinNT
        else begin
          if majorVer = 5 then
          begin
            if (majorVer = 5) and (minorVer = 0) then
              Result := cOsWin2000
            else if (majorVer = 5) and (minorVer = 1) then
              Result := cOsXP
            else
              Result := cOsUnknown;
          end;
        end;
      end;
    VER_PLATFORM_WIN32_WINDOWS: { Windows 9x/ME }
      begin
        if (majorVer = 4) and (minorVer = 0) then
          Result := cOsWin95
        else if (majorVer = 4) and (minorVer = 10) then
        begin
          if osVerInfo.szCSDVersion[1] = 'A' then
            Result := cOsWin98SE
          else
            Result := cOsWin98;
        end
        else
          if (majorVer = 4) and (minorVer = 90) then
            Result := cOsWinME
          else
            Result := cOsUnknown;
      end;
    else
      Result := cOsUnknown;
    end;
  end
  else
    Result := cOsUnknown;
end;

//--- Verify file is remote
function IsRemoteDrive(FileToCheck: string): Boolean;
var
  ReturnString: AnsiString;
  Size: DWORD;
  Drive: PWideChar;
begin
  try
    if ExtractFileDrive(FileToCheck) = '' then
      FileTocheck := ExtractFilePath(Application.ExeName) + FileToCheck;
    Result := (copy(FileToCheck, 1, 2) = '\\');
    if not Result then
      Result := (GetDriveType(PChar(ExtractFilePath(FileToCheck))) = DRIVE_REMOTE);
    if not Result then
    begin
      Size := 0;
      Drive := PWideChar(ExtractFileDrive(FileToCheck));
      if Win32Check(WNetGetConnection(Drive, nil, Size) in [NO_ERROR, ERROR_MORE_DATA]) then
      begin
        SetLength(ReturnString, Size);
        Win32Check(WNetGetConnection(Drive, PWideChar(ReturnString), Size) = NO_ERROR);
        SetLength(ReturnString, Size - 1);
        result := (copy(ReturnString, 1, 2) = '\\');
      end
      else
        result := false;
    end;
  except
    result := false;
  end;
end;

//--- Copy files from network to local application directory
function CopyHelpFiles(PatchNum, NetDir, LocalDir: string): Boolean;
var FileList: TStringList;
    fos : TSHFileOpStruct;
    i, iError: integer;
    tmpStr: string;
begin
  iError := 0;
  try
    //--- Get the list of network files to be copied to the local workstation
    tmpStr := StringReplace(fmtCheckName, '%s', '', [rfReplaceAll]);
    FileList := TStringList.Create;
    FileList.LoadFromFile(NetDir + tmpStr);
    //-- First two lines are a description of the file purpose and format
    FileList.Delete(0);
    FileList.Delete(0);
    FileList.Sort;

    //--- Delete all files and the local CCR application directory
    FillChar(fos, SizeOf(fos), 0);
    with fos do
    begin
      wFunc := FO_DELETE;
      pFrom := PChar(LocalDir+#0);       //double zero-terminated
      fFlags := FOF_SILENT or FOF_NOERRORUI or FOF_NOCONFIRMATION;
    end;
    ShFileOperation(fos);

    //--- Try to create or use base directory
    if not DirectoryExists(LocalDir) then
      if not ForceDirectories(LocalDir) then
        iError := 1;

    //--- Copy network files to CCR's local application directory
    for i := 0 to FileList.Count - 1 do
    begin
      tmpStr := piece(FileList[i], '^', 3);
      if not CopyFile(PChar(NetDir+tmpStr), PChar(LocalDir+tmpStr), false) then
        iError := iError + 1;
    end;

    //--- Rename the list of files text file to include the Patch number
    tmpStr := LocalDir + StringReplace(fmtCheckName, '%s', '', [rfReplaceAll]);
    if not SysUtils.RenameFile(tmpStr, LocalDir+Format(fmtCheckName, [PatchNum])) then
      iError := iError + 1;
  finally
    FileList.Clear;
    FreeAndNil(FileList);
  end;

  result := (iError = 0);
end;

//============= HELP THREAD =============

constructor tHelpThread.Create(PatchNum, LocalDir, NetDir, HelpName: string);
begin
   inherited Create(True);
   fLocalDir := LocalDir;
   fNetDir := NetDir;
   fPatchNum := PatchNum;
   fHelpName := Helpname;
   FreeOnTerminate := True;
end;

destructor tHelpThread.Destroy;
begin
   inherited;
   CoUninitialize;
end;

procedure tHelpThread.Execute;
begin
  if Terminated then
    Exit;

  CoInitialize(nil);
  if CopyHelpFiles(fPatchNum, fNetDir, fLocalDir) then
    Application.HelpFile := fLocalDir + fHelpName
  else
    MessageDlg508('', RSC01001, mtError, [mbOK], 0);

  Sleep(Random(100));
end;

end.
