unit uROR_Reports;

interface

uses
  Classes, uROR_Errors, uROR_Common, uROR_ListView, uROR_ParamsPanel,
  domimpl;

type
  TReportDescriptor = class;
  TReportParameters = class;

  //----------------------------- TAvailableReports ----------------------------

  TAvailableReports = class(TStringList)
  private
    fLoaded: Boolean;

    function  getObject(anIndex: Integer): TReportDescriptor; reintroduce;
    procedure setObject(anIndex: Integer; aValue: TReportDescriptor);

  public
    constructor Create;
    destructor  Destroy; override;


    function  Add(const aCode: String): Integer; override;
    procedure Clear; override;
    procedure Delete(anIndex: Integer); override;
    function  FindDescriptor(const aCode: String): TReportDescriptor;
    function  Load(const aRegistryIEN: String; Reload: Boolean = False): TReturnCode;

    property Objects[anIndex: Integer]: TReportDescriptor read    getObject
                                                          write   setObject;
                                                          default;

  end;

  //----------------------------- TReportDescriptor ----------------------------

  TReportDescriptor = class(TPersistent)
  private
    fCode:            String;
    fDefaultParams:   String;
    fDefaultSorting:  TStringList;
    fForeground:      Boolean;
    fName:            String;
    fPanels:          TStringList;
    fPanelTags:       String;
    fParameters:      TReportParameters;
    fParamsIEN:       String;
    fSharedTemplates: Boolean;

  public
    constructor Create;
    destructor  Destroy; override;

    procedure AssignRawData(aNode: TDomNode); virtual;
    procedure RestoreDefaultParams;

    property Code: String                      read fCode;
    property DefaultParams: String             read fDefaultParams;
    property DefaultSorting: TStringList       read fDefaultSorting;
    property Foreground: Boolean               read fForeground;
    property Name: String                      read fName;
    property Panels: TStringList               read fPanels;
    property PanelTags: String                 read fPanelTags;
    property Parameters: TReportParameters     read fParameters;
    property ParamsIEN: String                 read fParamsIEN;
    property SharedTemplates: Boolean          read fSharedTemplates;

  end;

  //----------------------------- TReportParameters ----------------------------

  TReportParameters = class(TCustomReportParameters)
  private
    fXML: TDomDocument;

    function  getText: String;
    procedure setText(const aValue: String);
    procedure setXML(aValue: TDomDocument);

  public
    constructor Create;
    destructor  Destroy; override;

    function  AppendNode(aParent: TDomNode; const aName: String;
      const aValue: String = ''): TDomNode;
    procedure Assign(aSource: TPersistent); override;
    procedure AssignNode(aNode: TDomNode);
    procedure AssignTo(aDest: TPersistent); override;
    function  DeleteChildren(aNode: TDomNode): TDomNode; overload;
    function  DeleteChildren(const aPath: String; aTop: TDomNode = nil): TDomNode; overload;
    procedure DeleteNode(aNode: TDomNode); overload;
    procedure DeleteNode(const aPath: String; aTop: TDomNode = nil); overload;
    function  FindNode(const aPath: String; aTop: TDomNode = nil): TDomNode;
    function  GetAttrBool(aNode: TDomNode; const aName: String): Boolean;
    function  GetAttrDouble(aNode: TDomNode; const aName: String): Double;
    function  GetAttrInt(aNode: TDomNode; const aName: String): Integer;
    function  GetAttrStr(aNode: TDomNode; const aName: String): String;
    function  GetNode(const aPath: String; aTop: TDomNode = nil): TDomNode;
    function  GetNodeVal(aNode: TDomNode): String; overload;
    function  GetNodeVal(const aPath: String; aTop: TDomNode = nil): String; overload;
    procedure SetAttrVal(aNode: TDomNode; const aName: String;
      const aValue: String); overload;
    procedure SetAttrVal(aNode: TDomNode; const aName: String;
      const aValue: Boolean); overload;
    function  SetNodeVal(const aPath: String; const aValue: String;
      aTop: TDomNode = nil): TDomNode; overload;
    function  SetNodeVal(const aPath: String; const aValue: Boolean;
      aTop: TDomNode = nil): TDomNode; overload;

    property Text: String read getText write setText;
    property XML: TDomDocument read fXML write setXML;

  end;

  procedure dumpNode(aNode: TDomNode; const aTitle: String);

implementation

uses
  SysUtils, uROR_CommonUtils, uROR_Broker, uROR_Utilities, StrUtils, uROR_XML,
  Dialogs;


procedure dumpNode(aNode: TDomNode; const aTitle: String);
var
  buf: String;
  log: TStringList;

  procedure process_node(node: TDomNode; level: Integer);
  var
    child, single_node: TDomNode;
    doc: TDomDocument;
  begin
    if Assigned(node) then
      begin
        doc := node.OwnerDocument;
        single_node := doc.cloneNode(True,node);
        buf := single_node.ToString;
        doc.DisposeNode(single_node);
        log.Add(DupeString(' ',level*2)+buf);
        child := node.FirstChild;
        while child <> nil do
          begin
            process_node(child, level+1);
            child := child.NextSibling;
          end;
      end;
  end;

begin
  log := TStringList.Create;
  try
    //log.Add(aTitle);
    //process_node(aNode,0);
    buf := aTitle + #13;
    buf := AnsiReplaceText(aNode.ToString, '><', '>'#13'<');
    log.Text := buf;
    MessageDlg508('', log.Text, mtInformation, [mbOK], 0);
  finally
    log.Free;
  end;
end;

/////////////////////////////// TAvailableReports \\\\\\\\\\\\\\\\\\\\\\\\\\\\\

constructor TAvailableReports.Create;
begin
  inherited;
  CaseSensitive := False;
  Duplicates    := dupError;
  fLoaded       := False;
  Sorted        := True;
end;

destructor TAvailableReports.Destroy;
begin
  Clear;
  inherited;
end;

function TAvailableReports.Add(const aCode: String): Integer;
var
  rptDesc: TReportDescriptor;
begin
  rptDesc := TReportDescriptor.Create;
  try
    Result := AddObject(aCode, rptDesc);
  except
    rptDesc.Free;
    Raise;
  end;
end;

procedure TAvailableReports.Clear;
var
  i: Integer;
begin
  for i:=Count-1 downto 0 do
    begin
      try
        Objects[i].Free;
      except
      end;
      Objects[i] := nil;
    end;
  inherited;
end;

procedure TAvailableReports.Delete(anIndex: Integer);
begin
  try
    Objects[anIndex].Free;
    Objects[anIndex] := nil;
  except
  end;
  inherited;
end;

function TAvailableReports.FindDescriptor(const aCode: String): TReportDescriptor;
var
  idx: Integer;
begin
  if Find(aCode, idx) then
    Result := Objects[idx]
  else
    Result := nil;
end;

function TAvailableReports.getObject(anIndex: Integer): TReportDescriptor;
begin
  Result := inherited Objects[anIndex] as TReportDescriptor;
end;

function TAvailableReports.Load(const aRegistryIEN: String; Reload: Boolean): TReturnCode;
var
  idx, irpn: Integer;
  rptCode: String;
  rawData: TStringList;
  xmlDoc: TDomDocument;
  rptNode: TDomNode;
  rptNodes: TDomNodeSet;
  {$IFDEF DBUGETM3}
  DebugETM: TextFile;
  {$ENDIF}

begin
  if not fLoaded or Reload then
    begin
      Clear;
      xmlDoc := nil;
      rptNodes := nil;
      rawData := TStringList.Create;
      try
        if Broker.CallProc(rpcReportsAvailable, [aRegistryIEN],
          nil, rawData) then
          begin
            {$IFDEF DBUGETM3}
            if FileExists('C:\ETM\DEBUG-ETM-3.txt') then
            begin
              System.AssignFile(DebugETM, 'C:\ETM\DEBUG-ETM-3.txt');
              System.Append(DebugETM);
              writeln(DebugETM, '******Start of TAvailableReports.Load******');
              writeln(DebugETM, 'Registry IEN=[' + aRegistryIEN + ']');
              writeln(DebugETM, 'rawData=[' + rawData.Text + ']');
              writeln(DebugETM, '******End of TAvailableReports.Load******');
              System.CloseFile(DebugETM);
            end;
            {$ENDIF}
            rawData.Delete(0);
            rawData.Strings[0] := '<?xml version="1.0"?> ';
            xmlDoc := XMLDOM.ParseBuffer(Utf8ToAnsi(rawData.Text));
            rawData.Text := xmlDoc.ToString;
            FreeAndNil(rawData);
            if Assigned(xmlDoc) then
              begin
                rptNodes := xmlDoc.Xql('/REPORTS/REPORT', xmlDoc);
                for irpn:=0 to rptNodes.Length-1 do
                  begin
                    rptNode := rptNodes.Item(irpn);
                    rptCode := rptNode.getAttribute('CODE');
                    if rptCode <> '' then
                      begin
                        idx := Add(rptCode);
                        Objects[idx].AssignRawData(rptNode);
                      end;
                  end;
              end;
            fLoaded := True;
            Result := RC_OK;
          end
        else
          Result := RCE_RPC_ERROR;
      finally
        rawData.Free;
        rptNodes.Free;
        if Assigned(xmlDoc) then
          XMLDOM.FreeDocument(xmlDoc);
      end;
    end
  else
    Result := RC_OK;
end;

procedure TAvailableReports.setObject(anIndex: Integer; aValue: TReportDescriptor);
begin
  inherited Objects[anIndex] := aValue;
end;

/////////////////////////////// TReportDescriptor \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

constructor TReportDescriptor.Create;
begin
  inherited;
  fDefaultSorting := TStringList.Create;
  fPanels         := TStringList.Create;
  fParameters     := TReportParameters.Create;
end;

destructor TReportDescriptor.Destroy;
begin
  FreeAndNil(fParameters);
  FreeAndNil(fPanels);
  FreeAndNil(fDefaultSorting);
  inherited;
end;

procedure TReportDescriptor.AssignRawData(aNode: TDomNode);
var
  node: TDomNode;
  tblName: String;
begin
  fName            := aNode.getAttribute('NAME');
  fForeground      := StrToBoolDef(aNode.getAttribute('FGP'), False);
  fCode            := aNode.getAttribute('CODE');
  fParamsIEN       := aNode.getAttribute('IEN');
  fSharedTemplates := StrToBoolDef(aNode.getAttribute('SHARED_TEMPLATES'), False);

  //--- List of panel tags
  node := aNode.FindNode('PANELS/text()');
  if Assigned(node) then
    fPanelTags := node.NodeValue
  else
    fPanelTags := '';

  //--- Assign default parameters
  node := aNode.FindNode('PARAMS');
  if Assigned(node) then
    fDefaultParams := node.ToString
  else
    fDefaultParams := '';
  Parameters.AssignNode(node);

  //--- Assign default sort modes
  DefaultSorting.Clear;
  node := aNode.FindNode('SORT_MODES');
  if Assigned(node) then
    begin
      node := node.FirstChild;
      while node <> nil do
        begin
          tblName := node.getAttribute('TABLE');
          if tblName <> '' then
            DefaultSorting.Values[tblName] := node.getAttribute('FIELD');
          node := node.NextSibling;
        end;
    end;
end;

procedure TReportDescriptor.RestoreDefaultParams;
begin
  Parameters.Text := DefaultParams;
end;

/////////////////////////////// TReportParameters \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

constructor TReportParameters.Create;
begin
  inherited;
  fXML := XMLDOM.CreateDocument;
end;

destructor TReportParameters.Destroy;
begin
  if Assigned(fXML) then
    begin
      XMLDOM.FreeDocument(fXML);
      fXML := nil;
    end;
  inherited;
end;

function TReportParameters.AppendNode(aParent: TDomNode;
  const aName, aValue: String):TDomNode;
begin
  Result := aParent.appendChild(XML.CreateElement(aName));
  if aValue <> '' then
    Result.appendChild(XML.CreateTextNode(aValue));
end;

procedure TReportParameters.Assign(aSource: TPersistent);
begin
  if Assigned(aSource) then
    if aSource is TStrings then
      XML := XMLDOM.ParseBuffer(Utf8ToAnsi(TStrings(aSource).Text))
    else
      inherited;
end;

procedure TReportParameters.AssignNode(aNode: TDomNode);
begin
  if Assigned(aNode) then
    XML := XMLDOM.ParseBuffer(Utf8ToAnsi(aNode.ToString))
  else
    XML := XMLDOM.CreateDocument;
end;

procedure TReportParameters.AssignTo(aDest: TPersistent);
begin
  if aDest is TStrings then
    TStrings(aDest).Text := AnsiReplaceStr(fXML.ToString, '><', '>'#13'<')
  else
    inherited;
end;

function TReportParameters.DeleteChildren(aNode: TDomNode): TDomNode;
var
  node, next: TDomNode;
begin
  Result := aNode;
  if aNode <> nil then
    begin
      next := aNode.FirstChild;
      while next <> nil do
        begin
          node := next;
          next := node.NextSibling;
          DeleteNode(node);
        end;
    end;
end;

function TReportParameters.DeleteChildren(const aPath: String;
  aTop: TDomNode = nil): TDomNode;
begin
  Result := DeleteChildren(FindNode(aPath, aTop));
end;

procedure TReportParameters.DeleteNode(aNode: TDomNode);
begin
  if aNode <> nil then
    begin
      if aNode.ParentNode <> nil then
        aNode.ParentNode.removeChild(aNode);
      XML.DisposeNode(aNode);
    end;
end;

procedure TReportParameters.DeleteNode(const aPath: String; aTop: TDomNode);
begin
  if aTop <> nil then
    DeleteNode(aTop.FindNode(aPath))
  else
    DeleteNode(XML.FindNode(aPath));
end;

function TReportParameters.FindNode(const aPath: String; aTop: TDomNode): TDomNode;
begin
  if aTop <> nil then
    Result := aTop.FindNode(aPath)
  else
    Result := XML.FindNode(aPath);
end;

function TReportParameters.GetAttrBool(aNode: TDomNode; const aName: String): Boolean;
begin
  if aNode <> nil then
    Result := StrToBoolDef(aNode.getAttribute(aName), False)
  else
    Result := False;
end;

function TReportParameters.GetAttrDouble(aNode: TDomNode; const aName: String): Double;
begin
  if aNode <> nil then
    Result := StrToFloatDef(aNode.getAttribute(aName), 0.0)
  else
    Result := 0.0;
end;

function TReportParameters.GetAttrInt(aNode: TDomNode; const aName: String): Integer;
begin
  if aNode <> nil then
    Result := StrToIntDef(aNode.getAttribute(aName), 0)
  else
    Result := 0;
end;

function TReportParameters.GetAttrStr(aNode: TDomNode; const aName: String): String;
begin
  if aNode <> nil then
    Result := aNode.getAttribute(aName)
  else
    Result := '';
end;

function TReportParameters.GetNode(const aPath: String; aTop: TDomNode): TDomNode;
var
  ip: Integer;
  nodeName: String;
  node: TDomNode;
begin
  if aTop <> nil then
    begin
      node := aTop;
      ip := 0;
    end
  else
    begin
      node := XML;
      ip := 1;
    end;

  //--- Try to find the node
  Result := node.FindNode(aPath);

  //--- If the node does not exist, create it
  if Result = nil then
    begin
      Result := node;
      while True do
        begin
          Inc(ip);
          nodeName := Piece(aPath, '/', ip);
          if nodeName = '' then Break;
          node := Result.FindNode(nodeName);
          if node = nil then
            Result := Result.appendChild(XML.CreateElement(Piece(nodeName, '[')))
          else
            Result := node;
        end;
    end;
end;

function TReportParameters.GetNodeVal(aNode: TDomNode): String;
var
  node: TDomNode;
begin
  Result := '';
  if aNode <> nil then
    begin
      node := aNode.FindNode('text()');
      if node <> nil then
        Result := node.NodeValue;
    end;
end;

function TReportParameters.GetNodeVal(const aPath: String; aTop: TDomNode = nil): String;
var
  node: TDomNode;
begin
  node := FindNode(aPath + '/text()');
  if node <> nil then
    Result := node.NodeValue
  else
    Result := '';
end;

function TReportParameters.getText: String;
begin
  Result := XML.ToString;
end;

procedure TReportParameters.SetAttrVal(aNode: TDomNode; const aName: String;
  const aValue: String);
begin
  if aValue <> '' then
    aNode.setAttribute(aName, aValue)
  else
    aNode.removeAttribute(aName);
end;

procedure TReportParameters.SetAttrVal(aNode: TDomNode; const aName: String;
  const aValue: Boolean);
begin
  if aValue then
    aNode.setAttribute(aName, '1')
  else
    aNode.removeAttribute(aName);
end;

function TReportParameters.SetNodeVal(const aPath: String;
  const aValue: String; aTop: TDomNode): TDomNode;
var
  txtNode: TDomNode;
begin
  Result := GetNode(aPath, aTop);
  txtNode := Result.FindNode('text()');
  if aValue <> '' then
    begin
      if txtNode = nil then
        Result.appendChild(XML.CreateTextNode(aValue))
      else
        txtNode.NodeValue := aValue;
    end
  else if txtNode <> nil then
    DeleteNode(txtNode);
end;

function TReportParameters.SetNodeVal(const aPath: String;
  const aValue: Boolean; aTop: TDomNode): TDomNode;
begin
  if aValue then
    Result := SetNodeVal(aPath, '1', aTop)
  else
    Result := SetNodeVal(aPath, '', aTop);
end;

procedure TReportParameters.setText(const aValue: String);
begin
  XML := XMLDOM.ParseBuffer(Utf8ToAnsi(aValue));
end;

procedure TReportParameters.setXML(aValue: TDomDocument);
begin
  if aValue <> fXML then
    begin
      if Assigned(fXML) then
        XMLDOM.FreeDocument(fXML);
      fXML := aValue;
    end;
end;

end.
